/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "audio_config.h"
#include "audio_output.h"

const double        audio_output_t::PLAYBACK_START_DELAY = 0.5;
const ASIOCallbacks audio_output_t::ASIO_CALLBACKS = { bufferSwitch,	sampleRateDidChange, asioMessage, bufferSwitchTimeInfo };
const uint8_t       audio_output_t::DoP_MARKER[2] = { 0x05, 0xFA };
audio_output_t*     audio_output_t::current_instance = nullptr;

bool audio_output_t::g_advanced_settings_query() {
	return false;
}

bool audio_output_t::g_needs_bitdepth_config() {
	return true;
}

bool audio_output_t::g_needs_dither_config() {
	return false;
}

bool audio_output_t::g_needs_device_list_prefixes() {
	return true;
}

void audio_output_t::g_enum_devices(output_device_enum_callback& p_callback) {
	g_audio_config.enum_devices(p_callback);
}

GUID audio_output_t::g_get_guid() {
	return audio_config_t::g_get_guid();
}

const char* audio_output_t::g_get_name() {
	return audio_config_t::g_get_name();
}

void audio_output_t::g_advanced_settings_popup(HWND p_parent, POINT p_menupoint) {
}

audio_output_t* audio_output_t::get_instance() {
	return current_instance;
}

audio_output_t::audio_output_t(const GUID& p_device, double p_buffer_length, bool p_dither, t_uint32 p_bitdepth) {
	current_instance = this;

	drv_mode = out_mode = STREAM_TYPE_PCM;
	run_mode = STREAM_TYPE_BAD;
	cmd_request_delayed_start = false;

	audio_buffer.any = nullptr;
	audio_cache_length = p_buffer_length;
	audio_stream->set_max_duration(p_buffer_length);
	
	pcm_bitmask = (-1) << (32 - p_bitdepth);
	scale = (audio_sample)1;
	fading = false;

	for (int i = 0; i < 256; i++) {
		swap_bits[i] = 0;
		for (int j = 0; j < 8; j++) {
			swap_bits[i] |= ((i >> j) & 1) << (7 - j);
		}
	}

	out_trace = audio_config_t::g_get_trace();

	if (out_trace) {
		pfc::print_guid str_device(p_device);
		pfc::string_formatter str_buffer_length;
		str_buffer_length << p_buffer_length;
		pfc::string_formatter str_dither;
		str_dither << (p_dither ? "Yes" : "No");
		console::printf(TRC_INF "asio_output_t::asio_output_t(device = {%s}, buffer_length = %s, dither = %s, bitdepth = %d)", (const char*)str_device, (const char*)str_buffer_length, (const char*)str_dither, p_bitdepth);
	}

	if (!asio_device.load(p_device)) {
		pfc::throw_exception_with_message<exception_io>("Cannot load audio device driver");
	}
}

audio_output_t::~audio_output_t() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::~asio_output_t(device = \"%s\")", asio_device.get_name());
	}
	cmd_stop();
	cmd_dispose();
	asio_device.unload();
	audio_stream->flush();
}

double audio_output_t::get_latency() {
	double duration = audio_stream->get_duration();
	/*
	if (out_trace) {
		pfc::string_formatter str_duration;
		str_duration << duration;
		console::printf(TRC_INF "asio_output_t::get_latency() => %s", (const char *)str_duration);
	}
	*/
	return duration;
}

void audio_output_t::process_samples(const audio_chunk& p_chunk) {
	t_samplespec spec;
	spec.fromchunk(p_chunk);
	/*
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::process_samples(channels = %d, samplerate = %d, samples = %d)", spec.m_channels, spec.m_sample_rate, p_chunk.get_sample_count());
	}
	*/
	if (!spec.is_valid()) {
		pfc::throw_exception_with_message<exception_io_data>("Invalid audio stream specifications");
	}
	stream_type_e ctl_mode;
	unsigned ctl_samplerate;
	audio_stream->get_control_params(spec.m_channels, spec.m_channel_config, ctl_mode, ctl_samplerate);
	if (ctl_mode == STREAM_TYPE_PCM) {
		audio_stream->write(p_chunk.get_data(), p_chunk.get_sample_count(), spec.m_channels, spec.m_sample_rate, spec.m_channel_config);
	}
}

void audio_output_t::update(bool& p_ready) {
	if (audio_stream->is_spec_change()) {
		t_samplespec spec = audio_stream->read_spec();
		if (spec.is_valid()) {
			if (out_trace) {
				console::printf(TRC_INF "asio_output_t::update(action = config, buffered_samples = %d)", audio_stream->get_samples());
			}
			cmd_stop();
			out_mode = audio_stream->read_type();
			if (out_spec != spec) {
				cmd_dispose();
				out_spec = spec;
				out_samplerate = spec.m_sample_rate;
				inp_channels = spec.m_channels;
				cmd_prepare();
			}
			audio_stream->reset_spec_change();
			cmd_request_delayed_start = true;
		}
		else {
			if (out_trace) {
				console::printf(TRC_INF "asio_output_t::update(action = stop, buffered_samples = %d)", audio_stream->get_samples());
			}
			cmd_stop();
			audio_stream->reset_spec_change();
		}
	}
	if (cmd_request_delayed_start) {
		if (audio_stream->get_duration() >= PLAYBACK_START_DELAY) {
			if (out_trace) {
				console::printf(TRC_INF "asio_output_t::update(action = start, buffered_samples = %d)", audio_stream->get_samples());
			}
			cmd_start();
			cmd_request_delayed_start = false;
		}
	}
	p_ready = (cmd_request_delayed_start || audio_stream->get_duration() < audio_cache_length);
}

void audio_output_t::pause(bool p_state) {
	if (p_state) {
		if (out_trace) {
			console::printf(TRC_INF "asio_output_t::pause(action = stop, buffered_samples = %d)", audio_stream->get_samples());
		}
		cmd_stop();
	}
	else {
		if (out_trace) {
			console::printf(TRC_INF "asio_output_t::pause(action = start, buffered_samples = %d)", audio_stream->get_samples());
		}
		cmd_start();
	}
}

void audio_output_t::flush() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::flush()");
	}
	audio_stream->flush();
}

void audio_output_t::flush_changing_track() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::flush_changing_track()");
	}
	audio_stream->flush();
}

void audio_output_t::force_play() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::force_play()");
	}
}

void audio_output_t::volume_set(double p_val) {
	if (out_trace) {
		pfc::string_formatter str_val;
		str_val << p_val;
		console::printf(TRC_INF "asio_output_t::volume_set(val = %s)", (const char*)str_val);
	}
	scale = audio_math::gain_to_scale(p_val);
}

bool audio_output_t::want_track_marks() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::want_track_marks()");
	}
	return true;
}

void audio_output_t::on_track_mark() {
	if (out_trace) {
		console::printf(TRC_INF "asio_output_t::on_track_mark()");
	}
}

void audio_output_t::enable_fading(bool p_state) {
	if (out_trace) {
		pfc::string_formatter str_state;
		str_state << (p_state ? "Yes" : "No");
		console::printf(TRC_INF "asio_output_t::enable_fading(state = %s)", (const char*)str_state);
	}
	fading = p_state;
}

/*
* Private methods
*/

unsigned audio_output_t::get_channel_index(unsigned p_config, unsigned p_channel) {
	if (inp_channels == 1) {
		if (p_channel & (audio_chunk::channel_front_left | audio_chunk::channel_front_center | audio_chunk::channel_front_right)) {
			return 0;
		}
		else {
			return ~0;
		}
	}
	for (unsigned index = 0; index < audio_chunk::defined_channel_count; index++) {
		if ((1 << index) & (p_config & p_channel)) {
			return index;
		}
	}
	return ~0;
}

void audio_output_t::reset_buffers() {
	dop_marker_n.set_count(0);
	out_buffer_size = 0;
	inp_buffer_size = 0;
	if (audio_buffer.any) {
		delete [] audio_buffer.any;
		audio_buffer.any = nullptr;
	}
}

bool audio_output_t::drv_set_state_prepared() {
	bool ok = false;
	if (!drv_set_channels_for_track()) {
		pfc::throw_exception_with_message<exception_io_data>("No available channels in audio stream");
	}
	if (out_mode == STREAM_TYPE_DSD || out_mode == STREAM_TYPE_DoP) {
		g_audio_config.load_DoP();
		out_mode = g_audio_config.get_DoP() ? STREAM_TYPE_DoP : STREAM_TYPE_DSD;
	}
	if (out_mode == STREAM_TYPE_DSD && drv_mode == STREAM_TYPE_PCM) {
		if (asio_device.set_run_DSD(true)) {
			drv_mode = STREAM_TYPE_DSD;
		}
		else {
			out_mode = STREAM_TYPE_DoP;
		}
	}
	if ((out_mode == STREAM_TYPE_PCM || out_mode == STREAM_TYPE_DoP) && drv_mode == STREAM_TYPE_DSD) {
		if (asio_device.set_run_DSD(false)) {
			drv_mode = STREAM_TYPE_PCM;
		}
	}
	drv_samplerate = ((out_mode == STREAM_TYPE_DoP) ? out_samplerate / 16 : out_samplerate);
	if (asio_device.set_samplerate(drv_samplerate)) {
		if (asio_device.get_samplerate() == drv_samplerate) {
			if (asio_device.create_buffers((ASIOCallbacks*)&ASIO_CALLBACKS)) {
				run_mode = out_mode;
				run_type = asio_device.get_channelInfos(run_mode == STREAM_TYPE_DSD)[0].type;
				run_bufferInfos = asio_device.get_bufferInfos();
				out_buffer_size = asio_device.get_samples();
				switch (out_mode) {
				case STREAM_TYPE_PCM:
					inp_buffer_size = out_buffer_size;
					audio_buffer.pcm = new audio_sample[inp_channels * inp_buffer_size];
					break;
				case STREAM_TYPE_DSD:
					inp_buffer_size = (out_buffer_size + 7) / 8;
					audio_buffer.dsd = new uint8_t[inp_channels * inp_buffer_size];
					break;
				case STREAM_TYPE_DoP:
					inp_buffer_size = 2 * out_buffer_size;
					audio_buffer.dsd = new uint8_t[inp_channels * inp_buffer_size];
					dop_marker_n.set_count(out_channels);
					for (t_size ch = 0; ch < out_channels; ch++) {
						dop_marker_n[ch] = 0;
					}
					break;
				}
				asio_device.get_latencies();
				demux_stream_to_buffers(1);
				ok = true;
			}
		}
	}
	return ok;
}

bool audio_output_t::drv_set_channels_for_track() {
	drv_channels = asio_device.get_channels();
	g_audio_config.load_channel_mapping(drv_channels);
	asio_device.get_bufferInfos().set_count(0);
	inp_channel_indexes.set_count(0);
	out_channels = 0;
	for (unsigned drv_channel = 0; drv_channel < drv_channels; drv_channel++) {
		unsigned inp_channel = g_audio_config.get_channel_mapping(drv_channel);
		unsigned inp_index = get_channel_index(out_spec.m_channel_config, inp_channel);
		if (inp_index != ~0) {
			asio_device.get_bufferInfos().increase_size(1);
			asio_device.get_bufferInfos()[out_channels].isInput = ASIOFalse;
			asio_device.get_bufferInfos()[out_channels].channelNum = drv_channel;
			inp_channel_indexes.increase_size(1);
			inp_channel_indexes[out_channels] = inp_index;
			out_channels++;
		}
	}
	return out_channels > 0;
}

bool audio_output_t::cmd_prepare() {
	bool ok = false;
	if (asio_device.get_state() == ASIO_STATE_INITIALIZED) {
		reset_buffers();
		if (!drv_set_state_prepared()) {
			pfc::throw_exception_with_message<exception_io>("Cannot prepare playback");
		}
	}
	return ok;
}

bool audio_output_t::cmd_dispose() {
	bool ok = false;
	if (asio_device.get_state() == ASIO_STATE_PREPARED) {
		if (asio_device.dispose_buffers()) {
			reset_buffers();
			ok = true;
		}
		else {
			pfc::throw_exception_with_message<exception_io>("Cannot dispose playback");
		}
	}
	return ok;
}

bool audio_output_t::cmd_start() {
	bool ok = false;
	if (asio_device.get_state() == ASIO_STATE_PREPARED) {
		if (asio_device.start()) {
			ok = true;
		}
		else {
			pfc::throw_exception_with_message<exception_io>("Cannot start playback");
		}
	}
	return ok;
}

bool audio_output_t::cmd_stop() {
	bool ok = false;
	if (asio_device.get_state() == ASIO_STATE_RUNNING) {
		if (asio_device.stop()) {
			ok = true;
		}
		else {
			pfc::throw_exception_with_message<exception_io>("Cannot stop playback");
		}
	}
	return ok;
}

void audio_output_t::change_endianness(uint8_t* p_data, t_size p_size, t_size p_sample_size) {
	for (t_size sample = 0; sample < p_size; sample++) {
		for (t_size n = 0; n < p_sample_size / 2; n++) {
			uint8_t temp = p_data[sample * p_sample_size + n];
			p_data[sample * p_sample_size + n] = p_data[sample * p_sample_size + p_sample_size - 1 - n];
			p_data[sample * p_sample_size + p_sample_size - 1 - n] = temp;
		}
	}
}

void audio_output_t::demux_stream_to_buffers(int p_double_buffer_index) {
	t_size read_samples = audio_stream->read(audio_buffer.any, inp_buffer_size);
	switch (run_mode) {
	case STREAM_TYPE_PCM:
		{
			audio_math::scale(audio_buffer.pcm, read_samples * inp_channels, audio_buffer.pcm, scale);
			for (unsigned out_channel = 0; out_channel < out_channels; out_channel++) {
				unsigned inp_channel_index = inp_channel_indexes[out_channel];
				for (unsigned sample = read_samples; sample < inp_buffer_size; sample++) {
					audio_buffer.pcm[sample * inp_channels + inp_channel_index] = (audio_sample)0;
				}
				uint8_t* out = static_cast<uint8_t*>(run_bufferInfos[out_channel].buffers[p_double_buffer_index]);
				t_size sampleSizeInBits = asio_util_t::getSampleSizeInBits(run_type);
				t_size sampleSizeInBytes = (sampleSizeInBits + 7) / 8;
				bool isFP = asio_util_t::isFP(run_type);
				bool isMSB = asio_util_t::isMSB(run_type);
				if (isFP) {
					if (sampleSizeInBits == 32) {
						for (unsigned sample = 0; sample < inp_buffer_size; sample++) {
							((float*)out)[sample] = (float)audio_buffer.pcm[sample * inp_channels + inp_channel_index];
						}
					}
					else if (sampleSizeInBits == 64) {
						for (unsigned sample = 0; sample < inp_buffer_size; sample++) {
							((double*)out)[sample] = (double)audio_buffer.pcm[sample * inp_channels + inp_channel_index];
						}
					}
				}
				else {
					audio_sample flt_to_int_scale = (audio_sample)((1 << (sampleSizeInBits - 1)) - 1);
					for (unsigned sample = 0; sample < inp_buffer_size; sample++) {
						audio_sample flt_value = audio_buffer.pcm[sample * inp_channels + inp_channel_index];
						if (flt_value > (audio_sample)+1) {
							flt_value = (audio_sample)+1;
						}
						else if (flt_value < (audio_sample)-1) {
							flt_value = (audio_sample)-1;
						}
						int int_value = (int)(flt_to_int_scale * flt_value + 0.5) & pcm_bitmask;
						memcpy(&out[sample * sampleSizeInBytes], &int_value, sampleSizeInBytes);
					}
				}
				if (isMSB) {
					change_endianness(out, out_buffer_size, sampleSizeInBytes);
				}
			}
		}
		break;
	case STREAM_TYPE_DSD:
		{
			for (unsigned out_channel = 0; out_channel < out_channels; out_channel++) {
				unsigned inp_channel_index = inp_channel_indexes[out_channel];
				for (t_size sample = read_samples; sample < inp_buffer_size; sample++) {
					audio_buffer.dsd[sample * inp_channels + inp_channel_index] = DSD_SILENCE_BYTE;
				}
				uint8_t* out = reinterpret_cast<uint8_t*>(run_bufferInfos[out_channel].buffers[p_double_buffer_index]);
				t_size sampleSizeInBits = asio_util_t::getSampleSizeInBits(run_type);
				bool isMSB = asio_util_t::isMSB(run_type);
				if (sampleSizeInBits == 1) {
					for (t_size sample = 0; sample < inp_buffer_size; sample++) {
						uint8_t value;
						value = audio_buffer.dsd[sample * inp_channels + inp_channel_index];
						out[sample] = isMSB ? value : swap_bits[value];
					}
				}
			}
		}
		break;
	case STREAM_TYPE_DoP:
		{
			for (unsigned out_channel = 0; out_channel < out_channels; out_channel++) {
				unsigned inp_channel_index = inp_channel_indexes[out_channel];
				for (unsigned sample = read_samples; sample < inp_buffer_size; sample++) {
					audio_buffer.dsd[sample * inp_channels + inp_channel_index] = DSD_SILENCE_BYTE;
				}
				uint8_t* out = static_cast<uint8_t*>(run_bufferInfos[out_channel].buffers[p_double_buffer_index]);
				t_size sampleSizeInBits = asio_util_t::getSampleSizeInBits(run_type);
				t_size sampleSizeInBytes = (sampleSizeInBits + 7) / 8;
				bool isMSB = asio_util_t::isMSB(run_type);
				unsigned o = 0;
				for (unsigned sample = 0; sample < inp_buffer_size; sample += 2) {
					uint8_t b_msb, b_lsb;
					b_msb = audio_buffer.dsd[(sample + 0) * inp_channels + inp_channel_index];
					b_lsb = audio_buffer.dsd[(sample + 1) * inp_channels + inp_channel_index];
					switch (sampleSizeInBytes) {
					case 4:
						out[o + 0] = 0x00;
						out[o + 1] = b_lsb;
						out[o + 2] = b_msb;
						out[o + 3] = DoP_MARKER[dop_marker_n[out_channel]];
						break;
					case 3:
						out[o + 0] = b_lsb;
						out[o + 1] = b_msb;
						out[o + 2] = DoP_MARKER[dop_marker_n[out_channel]];
						break;
					default:
						for (t_size n = 0; n < sampleSizeInBytes; n++) {
							out[o + n] = 0;
						}
						break;
					}
					o += sampleSizeInBytes;
					dop_marker_n[out_channel] = ++dop_marker_n[out_channel] & 1;
				}
				if (isMSB) {
					change_endianness(out, out_buffer_size, sampleSizeInBytes);
				}
			}
		}
		break;
	}
	/*
	asio_device.output_ready();
	*/
}

/*
* ASIO Callbacks
*/

void audio_output_t::bufferSwitch(long doubleBufferIndex, ASIOBool directProcess) {
	/*
	t_uint64 position;
	t_uint64 timestamp;
	get_instance()->asio_device.get_sample_position(position, timestamp);
	*/
	get_instance()->demux_stream_to_buffers(doubleBufferIndex);
}

ASIOTime* audio_output_t::bufferSwitchTimeInfo(ASIOTime* params, long doubleBufferIndex, ASIOBool directProcess) {
	ASIOTime* pTime = params;
	/*
	t_uint64 position;
	t_uint64 timestamp;
	position = ((uint64_t)params->timeInfo.samplePosition.hi << 32) | (uint64_t)params->timeInfo.samplePosition.lo;
	timestamp = ((uint64_t)params->timeInfo.systemTime.hi << 32) | (uint64_t)params->timeInfo.systemTime.lo;
	*/
	get_instance()->demux_stream_to_buffers(doubleBufferIndex);
	return pTime;
}

void audio_output_t::sampleRateDidChange(ASIOSampleRate sampleRate) {
}

long audio_output_t::asioMessage(long selector, long value, void* message, double* opt) {
	long rc;
	switch (selector) {
	case kAsioSelectorSupported:
		switch (value) {
		case kAsioSelectorSupported:
		case kAsioEngineVersion:
		case kAsioResetRequest:
		case kAsioBufferSizeChange:
		case kAsioResyncRequest:
		case kAsioLatenciesChanged:
		case kAsioSupportsTimeInfo:
		case kAsioSupportsTimeCode:
		case kAsioOverload:
			rc = TRUE;
			break;
		default:
			rc = FALSE;
			break;
		}
		break;
	case kAsioEngineVersion:
		rc = 2L;
		break;
	case kAsioResetRequest:
		console::printf(TRC_INF "Reset Request");
		rc = FALSE;
		break;
	case kAsioBufferSizeChange:
		console::printf(TRC_INF "Buffer Size Change");
		rc = FALSE;
		break;
	case kAsioResyncRequest:
		console::printf(TRC_INF "Resync Request");
		rc = FALSE;
		break;
	case kAsioLatenciesChanged:
		console::printf(TRC_INF "Latencies Changed");
		rc = FALSE;
		break;
	case kAsioSupportsTimeInfo:
		rc = TRUE;
		break;
	case kAsioSupportsTimeCode:
		rc = FALSE;
		break;
	case kAsioOverload:
		console::printf(TRC_INF "Overload");
		rc = TRUE;
		break;
	default:
		rc = FALSE;
		break;
	}
	return rc;
}

static output_factory_t<audio_output_t> g_output_sacd_factory;
