/*
* SACD Decoder plugin
* Copyright (c) 2011-2015 Maxim V.Anisiutkin <maxim.anisiutkin@gmail.com>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#pragma once

#include <foobar2000.h>

#define DSDxFs1   (44100 * 1)
#define DSDxFs2   (44100 * 2)
#define DSDxFs4   (44100 * 4)
#define DSDxFs8   (44100 * 8)
#define DSDxFs64  (44100 * 64)
#define DSDxFs128 (44100 * 128)
#define DSDxFs256 (44100 * 256)
#define DSDxFs512 (44100 * 512)

#define D48xFs1   (48000 * 1)
#define D48xFs2   (48000 * 2)
#define D48xFs4   (48000 * 4)
#define D48xFs8   (48000 * 8)
#define D48xFs64  (48000 * 64)
#define D48xFs128 (48000 * 128)
#define D48xFs256 (48000 * 256)
#define D48xFs512 (48000 * 512)

#define DSD_SAMPLERATE DSDxFs64
#define PCM_SAMPLERATE DSDxFs1

#define DSD_SILENCE_BYTE 0x69

enum stream_type_e { STREAM_TYPE_BAD = -1, STREAM_TYPE_PCM = 0, STREAM_TYPE_DSD = 1, STREAM_TYPE_DoP = 2 };

class audio_chunk_t {
	t_samplespec          m_spec;
	pfc::array_t<uint8_t> m_buffer;
	t_size                m_samples;
public:
	audio_chunk_t() {
		m_samples = 0;
	}
	
	audio_chunk_t(const audio_chunk_t& p_source) {
		m_spec = p_source.m_spec;
		m_buffer = p_source.m_buffer;
		m_samples = p_source.m_samples;
	}

	audio_chunk_t(const void* p_buffer, t_size p_samples, unsigned p_channels, unsigned p_samplerate, unsigned p_channel_config) {
		set_data(p_buffer, p_samples, p_channels, p_samplerate, p_channel_config);
	}

	stream_type_e read_type() {
		stream_type_e type;
		if (m_spec.m_sample_rate > 0) {
			if (m_spec.m_sample_rate < DSDxFs64) {
				type = STREAM_TYPE_PCM;
			}
			else {
				type = STREAM_TYPE_DSD;
			}
		}
		else {
			type = STREAM_TYPE_BAD;
		}
		return type;
	}

	t_samplespec get_spec() {
		return m_spec;
	}

	t_size get_channels() {
		return m_spec.m_channels;
	}

	t_size get_samples() {
		return m_samples;
	}

	t_size get_samplesize() {
		return (m_spec.m_sample_rate < DSDxFs64) ? sizeof(audio_sample) : sizeof(uint8_t);;
	}

	double get_duration() {
		double duration;
		if (m_spec.m_sample_rate > 0) {
			if (m_spec.m_sample_rate < DSDxFs64) {
				duration = (double)m_samples / (double)m_spec.m_sample_rate;
			}
			else {
				duration = (double)m_samples / (double)(m_spec.m_sample_rate / 8);
			}
		}
		else {
			duration = 0.0;
		}
		return duration;
	}

	void* get_data() {
		return m_buffer.get_ptr();
	}

	void set_data(const void* p_buffer, t_size p_samples, unsigned p_channels, unsigned p_samplerate, unsigned p_channel_config) {
		m_samples = p_samples;
		m_spec.m_channels = p_channels;
		m_spec.m_sample_rate = p_samplerate;
		m_spec.m_channel_config = p_channel_config;
		t_size buffer_size = m_samples * m_spec.m_channels * get_samplesize();
		m_buffer.append_fromptr(reinterpret_cast<const uint8_t*>(p_buffer), buffer_size);
	}
};

class NOVTABLE audio_stream_t : public service_base {
	FB2K_MAKE_SERVICE_INTERFACE_ENTRYPOINT(audio_stream_t);
public:
	virtual void get_control_params(unsigned p_channels, unsigned p_config, stream_type_e& p_type, unsigned& p_samplerate) = 0;
	virtual void set_control_params(stream_type_e p_type, unsigned p_samplerate, unsigned& p_channels, unsigned& p_config) = 0;
	virtual t_size get_samples() = 0;
	virtual double get_duration() = 0;
	virtual void set_max_duration(double p_duration) = 0;
	virtual double get_watermark() = 0;
	virtual t_samplespec read_spec() = 0;
	virtual bool is_spec_change() = 0;
	virtual void reset_spec_change() = 0;
	virtual stream_type_e read_type() = 0;
	virtual t_size read(void* p_buffer, t_size p_samples) = 0;
	virtual void write(const void* p_buffer, t_size p_samples, unsigned p_channels, unsigned p_samplerate, unsigned p_channel_config) = 0;
	virtual void flush() = 0;
};

class NOVTABLE audio_stream_impl : public audio_stream_t {
	pfc::chain_list_v2_t<audio_chunk_t*> m_chunks;
	CRITICAL_SECTION                     m_chunk_cs;
	t_size                               m_chunk_read_samples;
	t_size m_samples;
	double m_duration;
	double m_max_duration;
	t_samplespec m_last_read_spec;
public:
	audio_stream_impl();
	virtual ~audio_stream_impl();
	virtual void get_control_params(unsigned p_channels, unsigned p_channel_config, stream_type_e& p_type, unsigned& p_samplerate);
	virtual void set_control_params(stream_type_e p_type, unsigned p_samplerate, unsigned& p_channels, unsigned& p_channel_config);
	virtual t_size get_samples();
	virtual double get_duration();
	virtual void set_max_duration(double p_duration);
	virtual double get_watermark();
	virtual t_samplespec read_spec();
	virtual bool is_spec_change();
	virtual void reset_spec_change();
	virtual stream_type_e read_type();
	virtual t_size read(void* p_buffer, t_size p_samples);
	virtual void write(const void* p_buffer, t_size p_samples, unsigned p_channels, unsigned p_samplerate, unsigned p_channel_config);
	virtual void flush();
};
